@@ -80,6 +80,8 @@ struct pid_list {
 	int			free;
 } *pid_list;
 
+struct pid_list *comm_list;
+
 static unsigned int page_size;
 static int input_fd;
 static const char *default_input_file = "trace.dat";
@@ -334,7 +336,7 @@ static void add_filter(const char *filter, int neg)
 	filter_next = &ftr->next;
 }
 
-static void add_pid_filter(const char *arg)
+static void __add_filter(struct pid_list **head, const char *arg)
 {
 	struct pid_list *list;
 	char *pids = strdup(arg);
@@ -350,14 +352,24 @@ static void add_pid_filter(const char *arg)
 		list = malloc_or_die(sizeof(*list));
 		list->pid = pid;
 		list->free = free;
-		list->next = pid_list;
-		pid_list = list;
+		list->next = *head;
+		*head = list;
 		/* The first pid needs to be freed */
 		free = 0;
 		pid = strtok_r(NULL, ",", &sav);
 	}
 }
 
+static void add_comm_filter(const char *arg)
+{
+	__add_filter(&comm_list, arg);
+}
+
+static void add_pid_filter(const char *arg)
+{
+	__add_filter(&pid_list, arg);
+}
+
 static char *append_pid_filter(char *curr_filter, char *pid)
 {
 	char *filter;
@@ -390,11 +402,49 @@ static char *append_pid_filter(char *curr_filter, char *pid)
 	return filter;
 }
 
-static void make_pid_filter(void)
+static void convert_comm_filter(struct tracecmd_input *handle)
+{
+	struct pevent *pevent;
+	struct pid_list *list;
+	struct cmdline *cmdline;
+	char pidstr[100];
+
+	if (!comm_list)
+		return;
+
+	pevent = tracecmd_get_pevent(handle);
+
+	/* Seach for comm names and get their pids */
+	for (list = comm_list; list; list = list->next) {
+		cmdline = pevent_data_pid_from_comm(pevent, list->pid, NULL);
+		if (!cmdline) {
+			warning("comm: %s not in cmdline list", list->pid);
+			continue;
+		}
+		do {
+			sprintf(pidstr, "%d", pevent_cmdline_pid(pevent, cmdline));
+			add_pid_filter(pidstr);
+			cmdline = pevent_data_pid_from_comm(pevent, list->pid,
+							    cmdline);
+		} while (cmdline);
+	}
+
+	while (comm_list) {
+		list = comm_list;
+		comm_list = comm_list->next;
+		if (list->free)
+			free(list->pid);
+		free(list);
+	}
+}
+
+static void make_pid_filter(struct tracecmd_input *handle)
 {
 	struct pid_list *list;
 	char *str = NULL;
 
+	convert_comm_filter(handle);
+
 	if (!pid_list)
 		return;
 
@@ -421,12 +471,12 @@ static void process_filters(struct handle_list *handles)
 	struct filter *event_filter;
 	struct filter_str *filter;
 	struct pevent *pevent;
-	char *errstr;
+	char errstr[200];
 	int ret;
 
 	pevent = tracecmd_get_pevent(handles->handle);
 
-	make_pid_filter();
+	make_pid_filter(handles->handle);
 
 	while (filter_strings) {
 		filter = filter_strings;
@@ -439,13 +489,13 @@ static void process_filters(struct handle_list *handles)
 			die("malloc");
 
 		ret = pevent_filter_add_filter_str(event_filter->filter,
-						   filter->filter,
-						   &errstr);
-		if (ret < 0)
+						   filter->filter);
+		if (ret < 0) {
+			pevent_strerror(pevent, ret, errstr, sizeof(errstr));
 			die("Error filtering: %s\n%s",
 			    filter->filter, errstr);
+		}
 
-		free(errstr);
 		free(filter);
 
 		if (filter->neg) {
@@ -777,7 +827,6 @@ test_filters(struct filter *event_filters, struct pevent_record *record, int neg
 static struct pevent_record *
 get_next_record(struct handle_list *handles, int *next_cpu)
 {
-	unsigned long long ts;
 	struct pevent_record *record;
 	int found = 0;
 	int next;
@@ -792,7 +841,6 @@ get_next_record(struct handle_list *handles, int *next_cpu)
 
 	do {
 		next = -1;
-		ts = 0;
 		if (filter_cpus) {
 			long long last_stamp = -1;
 			struct pevent_record *precord;
@@ -1108,6 +1156,8 @@ static void set_event_flags(struct pevent *pevent, struct event_str *list,
 }
 
 enum {
+	OPT_event	= 246,
+	OPT_comm	= 247,
 	OPT_boundary	= 248,
 	OPT_stat	= 249,
 	OPT_pid		= 250,
@@ -1127,6 +1177,7 @@ void trace_report (int argc, char **argv)
 	struct event_str **raw_ptr = &raw_events;
 	struct event_str **nohandler_ptr = &nohandler_events;
 	const char *functions = NULL;
+	const char *print_event = NULL;
 	struct input_files *inputs;
 	struct handle_list *handles;
 	int show_stat = 0;
@@ -1162,9 +1213,11 @@ void trace_report (int argc, char **argv)
 		static struct option long_options[] = {
 			{"cpu", required_argument, NULL, OPT_cpu},
 			{"events", no_argument, NULL, OPT_events},
+			{"event", required_argument, NULL, OPT_event},
 			{"filter-test", no_argument, NULL, 'T'},
 			{"kallsyms", required_argument, NULL, OPT_kallsyms},
 			{"pid", required_argument, NULL, OPT_pid},
+			{"comm", required_argument, NULL, OPT_comm},
 			{"check-events", no_argument, NULL,
 				OPT_check_event_parsing},
 			{"nodate", no_argument, NULL, OPT_nodate},
@@ -1262,12 +1315,18 @@ void trace_report (int argc, char **argv)
 		case OPT_events:
 			print_events = 1;
 			break;
+		case OPT_event:
+			print_event = optarg;
+			break;
 		case OPT_kallsyms:
 			functions = optarg;
 			break;
 		case OPT_pid:
 			add_pid_filter(optarg);
 			break;
+		case OPT_comm:
+			add_comm_filter(optarg);
+			break;
 		case OPT_check_event_parsing:
 			check_event_parsing = 1;
 			break;
@@ -1342,7 +1401,12 @@ void trace_report (int argc, char **argv)
 		}
 
 		if (print_events) {
-			tracecmd_print_events(handle);
+			tracecmd_print_events(handle, NULL);
+			return;
+		}
+
+		if (print_event) {
+			tracecmd_print_events(handle, print_event);
 			return;
 		}
 
