@@ -463,42 +463,3 @@ void tracecmd_stop_recording(struct tracecmd_recorder *recorder)
 
 	recorder->stop = 1;
 }
-
-/**
- * tracecmd_stat_cpu - show the buffer stats of a particular CPU
- * @s: the trace_seq to record the data in.
- * @cpu: the CPU to stat
- *
- */
-void tracecmd_stat_cpu(struct trace_seq *s, int cpu)
-{
-	char buf[BUFSIZ];
-	char *tracing;
-	char *path;
-	int fd;
-	int r;
-
-	tracing = tracecmd_find_tracing_dir();
-	if (!tracing) {
-		errno = ENODEV;
-		goto out_free;
-	}
-
-	path = malloc_or_die(strlen(tracing) + 40);
-	if (!path)
-		goto out_free;
-
-	sprintf(path, "%s/per_cpu/cpu%d/stats", tracing, cpu);
-	fd = open(path, O_RDONLY);
-	if (fd < 0)
-		goto out_free;
-
-	while ((r = read(fd, buf, BUFSIZ)) > 0)
-		trace_seq_printf(s, "%.*s", r, buf);
-
-	close(fd);
-
- out_free:
-	free(path);
-	free(tracing);
-}
